###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from sqlalchemy.engine import create_engine
from sqlalchemy.engine.url import URL
from sqlalchemy.exc import SQLAlchemyError
import urllib

from vistrails.core.db.action import create_action
from vistrails.core.bundles.installbundle import install
from vistrails.core import debug
from vistrails.core.modules.config import ModuleSettings
from vistrails.core.modules.module_registry import get_module_registry
from vistrails.core.modules.vistrails_module import Module, ModuleError
from vistrails.core.upgradeworkflow import UpgradeWorkflowHandler
from vistrails.core.utils import versions_increasing

from vistrails.packages.tabledata.common import TableObject


class DBConnection(Module):
    """Connects to a database.

    If the URI you enter uses a driver which is not currently installed,
    VisTrails will try to set it up.
    """
    _input_ports = [('protocol', '(basic:String)'),
                    ('user', '(basic:String)',
                     {'optional': True}),
                    ('password', '(basic:String)',
                     {'optional': True}),
                    ('host', '(basic:String)',
                     {'optional': True}),
                    ('port', '(basic:Integer)',
                     {'optional': True}),
                    ('db_name', '(basic:String)')]
    _output_ports = [('connection', '(DBConnection)')]

    def compute(self):
        url = URL(drivername=self.get_input('protocol'),
                  username=self.force_get_input('user', None),
                  password=self.force_get_input('password', None),
                  host=self.force_get_input('host', None),
                  port=self.force_get_input('port', None),
                  database=self.get_input('db_name'))

        try:
            engine = create_engine(url)
        except ImportError, e:
            driver = url.drivername
            installed = False
            if driver == 'sqlite':
                raise ModuleError(self,
                                  "Python was built without sqlite3 support")
            elif (driver == 'mysql' or
                    driver == 'drizzle'): # drizzle is a variant of MySQL
                installed = install({
                        'pip': 'mysql-python',
                        'linux-debian': 'python-mysqldb',
                        'linux-ubuntu': 'python-mysqldb',
                        'linux-fedora': 'MySQL-python'})
            elif (driver == 'postgresql' or
                    driver == 'postgre'):   # deprecated alias
                installed = install({
                        'pip': 'psycopg2',
                        'linux-debian':'python-psycopg2',
                        'linux-ubuntu':'python-psycopg2',
                        'linux-fedora':'python-psycopg2'})
            elif driver == 'firebird':
                installed = install({
                        'pip': 'fdb',
                        'linux-fedora':'python-fdb'})
            elif driver == 'mssql' or driver == 'sybase':
                installed = install({
                        'pip': 'pyodbc',
                        'linux-debian':'python-pyodbc',
                        'linux-ubuntu':'python-pyodbc',
                        'linux-fedora':'pyodbc'})
            elif driver == 'oracle':
                installed = install({
                        'pip': 'cx_Oracle'})
            else:
                raise ModuleError(self,
                                  "SQLAlchemy couldn't connect: %s" %
                                  debug.format_exception(e))
            if not installed:
                raise ModuleError(self,
                                  "Failed to install required driver")
            try:
                engine = create_engine(url)
            except Exception, e:
                raise ModuleError(self,
                                  "Couldn't connect to the database: %s" %
                                  debug.format_exception(e))
        except SQLAlchemyError:
            # This is NoSuchModuleError in newer versions of SQLAlchemy but we
            # want compatibility here
            raise ModuleError(
                    self,
                    "SQLAlchemy has no support for protocol %r -- are you "
                    "sure you spelled that correctly?" % url.drivername)

        self.set_output('connection', engine.connect())


class SQLSource(Module):
    _settings = ModuleSettings(configure_widget=
            'vistrails.packages.sql.widgets:SQLSourceConfigurationWidget')
    _input_ports = [('connection', '(DBConnection)'),
                    ('cacheResults', '(basic:Boolean)'),
                    ('source', '(basic:String)')]
    _output_ports = [('result', '(org.vistrails.vistrails.tabledata:Table)'),
                     ('resultSet', '(basic:List)')]

    def is_cacheable(self):
        return False

    def compute(self):
        cached = False
        if self.has_input('cacheResults'):
            cached = self.get_input('cacheResults')
            self.is_cacheable = lambda: cached
        connection = self.get_input('connection')
        inputs = dict((k, self.get_input(k)) for k in self.inputPorts.iterkeys()
                  if k not in ('source', 'connection', 'cacheResults'))
        s = urllib.unquote(str(self.get_input('source')))

        try:
            transaction = connection.begin()
            results = connection.execute(s, inputs)
            try:
                rows = results.fetchall()
            except Exception:
                self.set_output('result', None)
                self.set_output('resultSet', None)
            else:
                # results.returns_rows is True
                # We don't use 'if return_rows' because this attribute didn't
                # use to exist
                table = TableObject.from_dicts(rows, results.keys())
                self.set_output('result', table)
                self.set_output('resultSet', rows)
            transaction.commit()
        except SQLAlchemyError, e:
            raise ModuleError(self, debug.format_exception(e))


_modules = [DBConnection, SQLSource]


def handle_module_upgrade_request(controller, module_id, pipeline):
    # Before 0.0.3, SQLSource's resultSet output was type ListOfElements (which
    #   doesn't exist anymore)
    # In 0.0.3, SQLSource's resultSet output was type List
    # In 0.1.0, SQLSource's output was renamed to result and is now a Table;
    #   this is totally incompatible and no upgrade code is possible
    #   the resultSet is kept for now for compatibility

    # Up to 0.0.4, DBConnection would ask for a password if one was necessary;
    #   this behavior has not been kept. There is now a password input port, to
    #   which you can connect a PasswordDialog from package dialogs if needed

    old_module = pipeline.modules[module_id]
    # DBConnection module from before 0.1.0: automatically add the password
    # prompt module
    if (old_module.name == 'DBConnection' and
            versions_increasing(old_module.version, '0.1.0')):
        reg = get_module_registry()
        # Creates the new module
        new_module = controller.create_module_from_descriptor(
                reg.get_descriptor(DBConnection))
        # Create the password module
        mod_desc = reg.get_descriptor_by_name(
                'org.vistrails.vistrails.dialogs', 'PasswordDialog')
        mod = controller.create_module_from_descriptor(mod_desc)
        # Adds a 'label' function to the password module
        ops = [('add', mod)]
        ops.extend(controller.update_function_ops(mod,
                                                  'label', ['Server password']))
        # Connects the password module to the new module
        conn = controller.create_connection(mod, 'result',
                                            new_module, 'password')
        ops.append(('add', conn))
        # Replaces the old module with the new one
        upgrade_actions = UpgradeWorkflowHandler.replace_generic(
                controller, pipeline,
                old_module, new_module,
                src_port_remap={'self': 'connection'})
        password_fix_action = create_action(ops)
        return upgrade_actions + [password_fix_action]

    return UpgradeWorkflowHandler.attempt_automatic_upgrade(
            controller, pipeline,
            module_id)


###############################################################################

import unittest


class TestSQL(unittest.TestCase):
    def test_query_sqlite3(self):
        """Queries a SQLite3 database.
        """
        import os
        import sqlite3
        import tempfile
        import urllib2
        from vistrails.tests.utils import execute, intercept_results
        identifier = 'org.vistrails.vistrails.sql'

        test_db_fd, test_db = tempfile.mkstemp(suffix='.sqlite3')
        os.close(test_db_fd)
        try:
            conn = sqlite3.connect(test_db)
            cur = conn.cursor()
            cur.execute('''
                    CREATE TABLE test(name VARCHAR(24) PRIMARY KEY,
                                      lastname VARCHAR(32) NOT NULL,
                                      age INTEGER NOT NULL)
                    ''')
            cur.executemany('''
                    INSERT INTO test(name, lastname, age)
                    VALUES(:name, :lastname, :age)
                    ''',
                    [{'name': 'John', 'lastname': 'Smith', 'age': 25},
                     {'name': 'Lara', 'lastname': 'Croft', 'age': 21}])
            conn.commit()
            conn.close()

            source = ('''
                    INSERT INTO test(name, lastname, age)
                    VALUES(:name, :lastname, :age)
                    ''')

            with intercept_results(DBConnection, 'connection', SQLSource, 'result') as (connection, table):
                self.assertFalse(execute([
                        ('DBConnection', identifier, [
                            ('protocol', [('String', 'sqlite')]),
                            ('db_name', [('String', test_db)]),
                        ]),
                        ('SQLSource', identifier, [
                            ('source', [('String', urllib2.quote(source))]),
                            ('name', [('String', 'Michael')]),
                            ('lastname', [('String', 'Buck')]),
                            ('age', [('Integer', '78')]),
                        ]),
                    ],
                    [
                        (0, 'connection', 1, 'connection'),
                    ],
                    add_port_specs=[
                        (1, 'input', 'name',
                         'org.vistrails.vistrails.basic:String'),
                        (1, 'input', 'lastname',
                         'org.vistrails.vistrails.basic:String'),
                        (1, 'input', 'age',
                         'org.vistrails.vistrails.basic:Integer'),
                    ]))

            self.assertEqual(len(connection), 1)
            connection[0].close()
            self.assertEqual(len(table), 1)
            self.assertIsNone(table[0])

            source = "SELECT name, lastname, age FROM test WHERE age > :age"

            with intercept_results(DBConnection, 'connection', SQLSource, 'result') as (connection, table):
                self.assertFalse(execute([
                        ('DBConnection', identifier, [
                            ('protocol', [('String', 'sqlite')]),
                            ('db_name', [('String', test_db)]),
                        ]),
                        ('SQLSource', identifier, [
                            ('source', [('String', urllib2.quote(source))]),
                            ('age', [('Integer', '22')]),
                        ]),
                    ],
                    [
                        (0, 'connection', 1, 'connection'),
                    ],
                    add_port_specs=[
                        (1, 'input', 'age',
                         'org.vistrails.vistrails.basic:Integer'),
                    ]))

            self.assertEqual(len(connection), 1)
            connection[0].close()
            self.assertEqual(len(table), 1)
            table, = table
            self.assertEqual(table.names, ['name', 'lastname', 'age'])
            self.assertEqual((table.rows, table.columns), (2, 3))
            self.assertEqual(set(table.get_column(1)),
                             set(['Smith', 'Buck']))
        finally:
            try:
                os.remove(test_db)
            except OSError:
                pass # Oops, we are leaking the file here...



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from PyQt4 import QtGui, QtCore
import os
import vistrails.api

CHOICE_OTHER_ALL = 0
CHOICE_OTHER = 1
CHOICE_RESOLVED = 2
CHOICE_OWN = 3
CHOICE_OWN_ALL = 4

class resolve_tags(QtGui.QWidget):
    def __init__(self, a, b, text, parent=None):
        QtGui.QWidget.__init__(self, parent)

        self.value = 4
        self.a = a
        self.b = b
        self.text = text

        self.initUI()

    def initUI(self):

        self.setWindowTitle("Tag conflict!")
        self.move(250, 200)

        # main layout
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        # info boxes
        hbox = QtGui.QHBoxLayout()
        vbox.addLayout(hbox)

        otherLabel = QtGui.QLabel(self)
        otherLabel.setAlignment(QtCore.Qt.AlignHCenter)
        otherLabel.setText("Other version")
        hbox.addWidget(otherLabel)

        ownLabel = QtGui.QLabel(self)
        ownLabel.setAlignment(QtCore.Qt.AlignHCenter)
        ownLabel.setText("Your version")
        hbox.addWidget(ownLabel)

        # tags
        self.tedit = QtGui.QLineEdit(self.text)
        self.tedit.setAlignment(QtCore.Qt.AlignHCenter)
        vbox.addWidget(self.tedit)

        # info boxes
        hbox = QtGui.QHBoxLayout()
        vbox.addLayout(hbox)

        # info boxes
        vbox1 = QtGui.QVBoxLayout()
        vbox1.setAlignment(QtCore.Qt.AlignHCenter)
        hbox.addLayout(vbox1)
        userLabel = QtGui.QLabel(self)
        userLabel.setText("User: " + self.a.db_user)
        vbox1.addWidget(userLabel)
        timeLabel = QtGui.QLabel(self)
        timeLabel.setText("Time: " + str(self.a.db_date))
        vbox1.addWidget(timeLabel)

        vbox2 = QtGui.QVBoxLayout()
        vbox2.setAlignment(QtCore.Qt.AlignHCenter)
        hbox.addLayout(vbox2)
        userLabel = QtGui.QLabel(self)
        userLabel.setText("User: " + self.b.db_user)
        vbox2.addWidget(userLabel)
        timeLabel = QtGui.QLabel(self)
        timeLabel.setText("Time: " + str(self.b.db_date))
        vbox2.addWidget(timeLabel)

        buttons = QtGui.QHBoxLayout()
        vbox.addLayout(buttons)

        self.otherAll = QtGui.QPushButton('Keep other (for all)', self)
        self.connect(self.otherAll, QtCore.SIGNAL('clicked()'), self.setOtherAll)
        buttons.addWidget(self.otherAll)
        self.other = QtGui.QPushButton('Keep other', self)
        self.connect(self.other, QtCore.SIGNAL('clicked()'), self.setOther)
        buttons.addWidget(self.other)
        self.resolve = QtGui.QPushButton('Resolved', self)
        self.connect(self.resolve, QtCore.SIGNAL('clicked()'), self.setResolved)
        buttons.addWidget(self.resolve)
        self.own = QtGui.QPushButton('Keep yours', self)
        self.connect(self.own, QtCore.SIGNAL('clicked()'), self.setOwn)
        buttons.addWidget(self.own)
        self.ownAll = QtGui.QPushButton('Keep yours (for all)', self)
        self.connect(self.ownAll, QtCore.SIGNAL('clicked()'), self.setOwnAll)
        buttons.addWidget(self.ownAll)

    def setOtherAll(self):
        self.value = CHOICE_OTHER_ALL
        self.close()
    def setOther(self):
        self.value = CHOICE_OTHER
        self.close()
    def setResolved(self):
        self.value = CHOICE_RESOLVED
        self.text = str(self.tedit.text())
        self.close()
    def setOwn(self):
        self.value = CHOICE_OWN
        self.close()
    def setOwnAll(self):
        self.value = CHOICE_OWN_ALL
        self.close()

class resolve_notes(QtGui.QWidget):
    def __init__(self, a, b, text, parent=None):
        QtGui.QWidget.__init__(self, parent)

        self.value = 4
        self.a = a
        self.b = b
        self.text = text

        self.initUI()

    def initUI(self):

        self.setWindowTitle("Notes conflict!")
        self.move(250, 200)

        # main layout
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        # thumbs layout
        self.tedit = QtGui.QTextEdit(self.text)
        vbox.addWidget(self.tedit)

        buttons = QtGui.QHBoxLayout()
        vbox.addLayout(buttons)

        self.otherAll = QtGui.QPushButton('Keep other (for all)', self)
        self.connect(self.otherAll, QtCore.SIGNAL('clicked()'), self.setOtherAll)
        buttons.addWidget(self.otherAll)
        self.other = QtGui.QPushButton('Keep other', self)
        self.connect(self.other, QtCore.SIGNAL('clicked()'), self.setOther)
        buttons.addWidget(self.other)
        self.resolve = QtGui.QPushButton('Resolved', self)
        self.connect(self.resolve, QtCore.SIGNAL('clicked()'), self.setResolved)
        buttons.addWidget(self.resolve)
        self.own = QtGui.QPushButton('Keep yours', self)
        self.connect(self.own, QtCore.SIGNAL('clicked()'), self.setOwn)
        buttons.addWidget(self.own)
        self.ownAll = QtGui.QPushButton('Keep yours (for all)', self)
        self.connect(self.ownAll, QtCore.SIGNAL('clicked()'), self.setOwnAll)
        buttons.addWidget(self.ownAll)

    def setOtherAll(self):
        self.value = CHOICE_OTHER_ALL
        self.close()
    def setOther(self):
        self.value = CHOICE_OTHER
        self.close()
    def setResolved(self):
        self.value = CHOICE_RESOLVED
        self.text = str(self.tedit.toHtml())
        self.close()
    def setOwn(self):
        self.value = CHOICE_OWN
        self.close()
    def setOwnAll(self):
        self.value = CHOICE_OWN_ALL
        self.close()

class resolve_thumbs(QtGui.QWidget):
    def __init__(self, a, b, old_tmp_dir, new_tmp_dir, parent=None):
        QtGui.QWidget.__init__(self, parent)

        self.value = 4
        self.a = a
        self.b = b

        self.old_tmp_dir = old_tmp_dir
        self.new_tmp_dir = new_tmp_dir
        self.initUI()

    def initUI(self):

        self.setWindowTitle("Thumbnail conflict!")
        self.move(250, 200)

        # main layout
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        # thumbs layout
        hbox = QtGui.QHBoxLayout()
        vbox.addLayout(hbox)

        other = QtGui.QVBoxLayout()
        hbox.addLayout(other)
        otherLabel = QtGui.QLabel(self)
        otherLabel.setText("Other version")
        other.addWidget(otherLabel)
        path = os.path.join(self.old_tmp_dir, 'thumbs', self.a.db_value)
        pixmap = QtGui.QPixmap(path)
        thumb1 = QtGui.QLabel(self)
        thumb1.setPixmap(pixmap)
        other.addWidget(thumb1)
        userLabel = QtGui.QLabel(self)
        userLabel.setText("User: " + self.a.db_user)
        other.addWidget(userLabel)
        timeLabel = QtGui.QLabel(self)
        timeLabel.setText("Time: " + str(self.a.db_date))
        other.addWidget(timeLabel)

        own = QtGui.QVBoxLayout()
        hbox.addLayout(own)
        ownLabel = QtGui.QLabel(self)
        ownLabel.setText("Your version")
        own.addWidget(ownLabel)
        path = os.path.join(self.old_tmp_dir, 'thumbs', self.b.db_value)
        pixmap = QtGui.QPixmap(path)
        thumb2 = QtGui.QLabel(self)
        thumb2.setPixmap(pixmap)
        own.addWidget(thumb2)
        userLabel = QtGui.QLabel(self)
        userLabel.setText("User: " + self.b.db_user)
        own.addWidget(userLabel)
        timeLabel = QtGui.QLabel(self)
        timeLabel.setText("Time: " + str(self.b.db_date))
        own.addWidget(timeLabel)

        buttons = QtGui.QHBoxLayout()
        vbox.addLayout(buttons)

        self.otherAll = QtGui.QPushButton('Keep other (for all)', self)
        self.connect(self.otherAll, QtCore.SIGNAL('clicked()'), self.setOtherAll)
        buttons.addWidget(self.otherAll)
        self.other = QtGui.QPushButton('Keep other', self)
        self.connect(self.other, QtCore.SIGNAL('clicked()'), self.setOther)
        buttons.addWidget(self.other)
        self.own = QtGui.QPushButton('Keep yours', self)
        self.connect(self.own, QtCore.SIGNAL('clicked()'), self.setOwn)
        buttons.addWidget(self.own)
        self.ownAll = QtGui.QPushButton('Keep yours (for all)', self)
        self.connect(self.ownAll, QtCore.SIGNAL('clicked()'), self.setOwnAll)
        buttons.addWidget(self.ownAll)

    def setOtherAll(self):
        self.value = CHOICE_OTHER_ALL
        self.close()
    def setOther(self):
        self.value = CHOICE_OTHER
        self.close()
    def setOwn(self):
        self.value = CHOICE_OWN
        self.close()
    def setOwnAll(self):
        self.value = CHOICE_OWN_ALL
        self.close()

class MergeGUI(object):
    @staticmethod
    def resolveTags(a, b, text):
        exm = resolve_tags(a, b, text)
        exm.show()
        app = vistrails.api.get_builder_window()
        if not app:
            app = QtGui.QApplication([])
            app.exec_()
        return exm.value, exm.text

    @staticmethod
    def resolveNotes(a, b, text):
        exm = resolve_notes(a, b, text)
        exm.show()
        app = vistrails.api.get_builder_window()
        if not app:
            app = QtGui.QApplication([])
            app.exec_()
        return exm.value, exm.text

    @staticmethod
    def resolveThumbs(a, b, old_tmp_dir, new_tmp_dir):
        exm = resolve_thumbs(a, b, old_tmp_dir, new_tmp_dir)
        exm.show()
        app = vistrails.api.get_builder_window()
        if not app:
            app = QtGui.QApplication([])
            app.exec_()
        return exm.value



#!/usr/bin/env python

# This example shows how to use cutting (vtkCutter) and how it
# compares with extracting a plane from a computational grid.

import vtk
from vtk.util.misc import vtkGetDataRoot
VTK_DATA_ROOT = vtkGetDataRoot()

# Read some data.
pl3d = vtk.vtkPLOT3DReader()
pl3d.SetXYZFileName(VTK_DATA_ROOT + "/Data/combxyz.bin")
pl3d.SetQFileName(VTK_DATA_ROOT + "/Data/combq.bin")
pl3d.SetScalarFunctionNumber(100)
pl3d.SetVectorFunctionNumber(202)
pl3d.Update()

# The cutter uses an implicit function to perform the cutting.
# Here we define a plane, specifying its center and normal.
# Then we assign the plane to the cutter.
plane = vtk.vtkPlane()
plane.SetOrigin(pl3d.GetOutput().GetCenter())
plane.SetNormal(-0.287, 0, 0.9579)
planeCut = vtk.vtkCutter()
planeCut.SetInputConnection(pl3d.GetOutputPort())
planeCut.SetCutFunction(plane)
cutMapper = vtk.vtkPolyDataMapper()
cutMapper.SetInputConnection(planeCut.GetOutputPort())
cutMapper.SetScalarRange(pl3d.GetOutput().GetPointData().GetScalars().GetRange())
cutActor = vtk.vtkActor()
cutActor.SetMapper(cutMapper)

# Here we extract a computational plane from the structured grid.
# We render it as wireframe.
compPlane = vtk.vtkStructuredGridGeometryFilter()
compPlane.SetInputConnection(pl3d.GetOutputPort())
compPlane.SetExtent(0, 100, 0, 100, 9, 9)
planeMapper = vtk.vtkPolyDataMapper()
planeMapper.SetInputConnection(compPlane.GetOutputPort())
planeMapper.ScalarVisibilityOff()
planeActor = vtk.vtkActor()
planeActor.SetMapper(planeMapper)
planeActor.GetProperty().SetRepresentationToWireframe()
planeActor.GetProperty().SetColor(0, 0, 0)

# The outline of the data puts the data in context.
outline = vtk.vtkStructuredGridOutlineFilter()
outline.SetInputConnection(pl3d.GetOutputPort())
outlineMapper = vtk.vtkPolyDataMapper()
outlineMapper.SetInputConnection(outline.GetOutputPort())
outlineActor = vtk.vtkActor()
outlineActor.SetMapper(outlineMapper)
outlineProp = outlineActor.GetProperty()
outlineProp.SetColor(0, 0, 0)

# Create the RenderWindow, Renderer and both Actors
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Add the actors to the renderer, set the background and size
ren.AddActor(outlineActor)
ren.AddActor(planeActor)
ren.AddActor(cutActor)

ren.SetBackground(1, 1, 1)
renWin.SetSize(400, 300)

cam1 = ren.GetActiveCamera()
cam1.SetClippingRange(11.1034, 59.5328)
cam1.SetFocalPoint(9.71821, 0.458166, 29.3999)
cam1.SetPosition(-2.95748, -26.7271, 44.5309)
cam1.SetViewUp(0.0184785, 0.479657, 0.877262)

iren.Initialize()
renWin.Render()
iren.Start()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This modules builds a widget to interact with vistrail diff
operation """
from __future__ import division

from PyQt4 import QtCore, QtGui
from vistrails.core.system import get_vistrails_basic_pkg_id
from vistrails.core.utils import VistrailsInternalError
from vistrails.core.utils.color import ColorByName
from vistrails.core.vistrail.abstraction import Abstraction
from vistrails.core.vistrail.pipeline import Pipeline
from vistrails.core.vistrail.port_spec import PortSpec
from vistrails.gui.pipeline_view import QPipelineView
from vistrails.gui.theme import CurrentTheme
from vistrails.gui.vistrail_controller import VistrailController
from vistrails.gui.vistrails_palette import QVistrailsPaletteInterface
from vistrails.core import system, debug
import vistrails.core.db.io

import copy
from itertools import chain

################################################################################

class QFunctionItemModel(QtGui.QStandardItemModel):
    """
    QFunctionItemModel is a item model that will allow item to be
    show as a disabled one on the table
    
    """
    def __init__(self, row, col, parent=None):
        """ QFunctionItemModel(row: int, col: int, parent: QWidget)
                              -> QFunctionItemModel                             
        Initialize with a number of rows and columns
        
        """
        QtGui.QStandardItemModel.__init__(self, row, col, parent)
        self.disabledRows = {}

    def flags(self, index):
        """ flags(index: QModelIndex) -> None
        Return the current flags of the item with the index 'index'
        
        """
        if index.isValid() and self.disabledRows.has_key(index.row()):
            return (QtCore.Qt.ItemIsDragEnabled | QtCore.Qt.ItemIsDropEnabled |
                    QtCore.Qt.ItemIsSelectable)
        return QtGui.QStandardItemModel.flags(self,index)

    def clearList(self):
        """ clearList() -> None
        Clear all items including the disabled ones
        
        """
        self.disabledRows = {}
        self.removeRows(0,self.rowCount())

    def disableRow(self,row):
        """ disableRow(row: int) -> None
        Disable a specific row on the table
        
        """
        self.disabledRows[row] = None

class QParamTable(QtGui.QTableView):
    """
    QParamTable is a widget represents a diff between two version
    as side-by-side comparisons
    
    """
    def __init__(self, v1Name=None, v2Name=None, parent=None):
        """ QParamTable(v1Name: str, v2Name: str, parent: QWidget)
                       -> QParamTable
        Initialize the table with two version names on the header view
        
        """
        QtGui.QTableView.__init__(self, parent)
        itemModel = QFunctionItemModel(0, 2, self)
        itemModel.setHeaderData(0, QtCore.Qt.Horizontal, v1Name)
        itemModel.setHeaderData(1, QtCore.Qt.Horizontal, v2Name)
        # self.setHorizontalHeaderLabels([v1Name, v2Name])
        self.setModel(itemModel)
        self.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)        
        self.setSelectionMode(QtGui.QAbstractItemView.SingleSelection)        
        self.setEditTriggers(QtGui.QAbstractItemView.NoEditTriggers)        
        self.setFont(CurrentTheme.VISUAL_DIFF_PARAMETER_FONT)
        self.horizontalHeader().setResizeMode(QtGui.QHeaderView.Stretch)
        
    def set_names(self, v1_name, v2_name):
        self.model().setHeaderData(0, QtCore.Qt.Horizontal, v1_name)
        self.model().setHeaderData(1, QtCore.Qt.Horizontal, v2_name)
        # self.setHorizontalHeaderLabels([v1_name, v2_name])

class QParamInspector(QtGui.QWidget):
    """
    QParamInspector is a widget acting as an inspector vistrail modules
    in diff mode. It consists of a function inspector and annotation
    inspector
    
    """
    def __init__(self, v1Name='', v2Name='',
                 parent=None, f=QtCore.Qt.WindowFlags()):
        """ QParamInspector(v1Name: str, v2Name: str,
                            parent: QWidget, f: WindowFlags)
                            -> QParamInspector
        Construct a widget containing tabs: Functions and Annotations,
        each of them is in turn a table of two columns for two
        corresponding versions.
        
        """
        QtGui.QWidget.__init__(self, parent, f | QtCore.Qt.Tool)
        self.setWindowTitle('Parameter Inspector - None')
        self.firstTime = True        
        self.boxLayout = QtGui.QVBoxLayout()
        self.boxLayout.setMargin(0)
        self.boxLayout.setSpacing(0)
        self.tabWidget = QtGui.QTabWidget()
        self.tabWidget.setTabPosition(QtGui.QTabWidget.North)
        self.tabWidget.setTabShape(QtGui.QTabWidget.Triangular)
        self.functionsTab = QParamTable(v1Name, v2Name)
        self.tabWidget.addTab(self.functionsTab, 'Functions')        
# FIXME add annotation support back in
#         self.annotationsTab = QParamTable(v1Name, v2Name)
#         self.annotationsTab.horizontalHeader().setStretchLastSection(True)
#         self.tabWidget.addTab(self.annotationsTab, 'Annotations')        
        self.boxLayout.addWidget(self.tabWidget)
        sizeGrip = QtGui.QSizeGrip(self)
        self.boxLayout.addWidget(sizeGrip)
        self.boxLayout.setAlignment(sizeGrip, QtCore.Qt.AlignRight)
        self.setLayout(self.boxLayout)

    def closeEvent(self, e):
        """ closeEvent(e: QCloseEvent) -> None        
        Doesn't allow the QParamInspector widget to close, but just hide
        instead
        
        """
        e.ignore()
        self.parent().showInspectorAction.setChecked(False)
        

class QLegendBox(QtGui.QFrame):
    """
    QLegendBox is just a rectangular box with a solid color
    
    """
    def __init__(self, brush, size, parent=None, f=QtCore.Qt.WindowFlags()):
        """ QLegendBox(color: QBrush, size: (int,int), parent: QWidget,
                      f: WindowFlags) -> QLegendBox
        Initialize the widget with a color and fixed size
        
        """
        QtGui.QFrame.__init__(self, parent, f)
        self.setFrameStyle(QtGui.QFrame.Box | QtGui.QFrame.Plain)
        self.setAttribute(QtCore.Qt.WA_PaintOnScreen)
        self.setAutoFillBackground(True)
        palette = QtGui.QPalette(self.palette())
        palette.setBrush(QtGui.QPalette.Window, brush)
        self.setPalette(palette)
        self.setFixedSize(*size)
        if system.systemType in ['Darwin']:
            #the mac's nice looking messes up with the colors
            if QtCore.QT_VERSION < 0x40500:
                self.setAttribute(QtCore.Qt.WA_MacMetalStyle, False)
            else:
                self.setAttribute(QtCore.Qt.WA_MacBrushedMetal, False)
        

class QLegendWindow(QtGui.QWidget):
    """
    QLegendWindow contains a list of QLegendBox and its description
    
    """
    def __init__(self, v1Name='', v2Name='', parent=None,
                 f=QtCore.Qt.WindowFlags()):
        """ QLegendWindow(v1Name: str, v2Name: str,
                          parent: QWidget, f: WindowFlags)
                          -> QLegendWindow
        Construct a window by default with 4 QLegendBox and 4 QLabels
        
        """
        QtGui.QWidget.__init__(self, parent, f | QtCore.Qt.Tool)
        self.setWindowTitle('Visual Diff Legend')
        self.firstTime = True
        self.gridLayout = QtGui.QGridLayout(self)
        self.gridLayout.setMargin(10)
        self.gridLayout.setSpacing(10)
        self.setFont(CurrentTheme.VISUAL_DIFF_LEGEND_FONT)
        
        parent = self.parent()
        
        self.legendV1Box = QLegendBox(
            CurrentTheme.VISUAL_DIFF_FROM_VERSION_BRUSH,
            CurrentTheme.VISUAL_DIFF_LEGEND_SIZE,
            self)        
        self.gridLayout.addWidget(self.legendV1Box, 0, 0)
        self.legendV1 = QtGui.QLabel(v1Name, self)
        self.gridLayout.addWidget(self.legendV1, 0, 1)
        
        self.legendV2Box = QLegendBox(
            CurrentTheme.VISUAL_DIFF_TO_VERSION_BRUSH,            
            CurrentTheme.VISUAL_DIFF_LEGEND_SIZE,
            self)        
        self.gridLayout.addWidget(self.legendV2Box, 1, 0)
        self.legendV2 = QtGui.QLabel(v2Name, self)
        self.gridLayout.addWidget(self.legendV2, 1, 1)
        
        self.legendV12Box = QLegendBox(CurrentTheme.VISUAL_DIFF_SHARED_BRUSH,
                                       CurrentTheme.VISUAL_DIFF_LEGEND_SIZE,
                                       self)
        self.gridLayout.addWidget(self.legendV12Box, 2, 0)
        self.legendV12 = QtGui.QLabel("Shared", self)
        self.gridLayout.addWidget(self.legendV12, 2, 1)
        
        self.legendParamBox = QLegendBox(
            CurrentTheme.VISUAL_DIFF_PARAMETER_CHANGED_BRUSH,
            CurrentTheme.VISUAL_DIFF_LEGEND_SIZE,
            self)
        self.gridLayout.addWidget(self.legendParamBox,3,0)
        self.legendParam = QtGui.QLabel("Parameter Changes", self)
        self.gridLayout.addWidget(self.legendParam,3,1)

        # self.legendMatchedBox = \
        #     QLegendBox(CurrentTheme.VISUAL_DIFF_MATCH_BRUSH,
        #                CurrentTheme.VISUAL_DIFF_LEGEND_SIZE,
        #                self)
        # self.gridLayout.addWidget(self.legendMatchedBox, 4, 0)
        # self.legendMatched = QtGui.QLabel("Matched", self)
        # self.gridLayout.addWidget(self.legendMatched, 4, 1)

        self.adjustSize()
        
    def set_names(self, v1_name, v2_name):
        self.legendV1.setText(v1_name)
        self.legendV2.setText(v2_name)

    def closeEvent(self,e):
        """ closeEvent(e: QCloseEvent) -> None
        Doesn't allow the Legend widget to close, but just hide
        instead
        
        """
        e.ignore()
        self.parent().showLegendsAction.setChecked(False)
        
class QDiffProperties(QtGui.QWidget, QVistrailsPaletteInterface):
    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self, parent)

        self.controller = None
        self.set_title("Diff Properties")
 
        layout = QtGui.QVBoxLayout()
        layout.setMargin(0)
        layout.setSpacing(0)
        self.legend = QLegendWindow()
        legend_group = QtGui.QGroupBox("Legend")
        g_layout = QtGui.QVBoxLayout()
        g_layout.setMargin(0)
        g_layout.setSpacing(0)
        g_layout.addWidget(self.legend)
        legend_group.setLayout(g_layout)
        layout.addWidget(legend_group)
        layout.setStretch(0,0)
        layout.addStretch(1)
        self.params = QParamTable()
        params_group = QtGui.QGroupBox("Parameter Changes")
        g_layout = QtGui.QVBoxLayout()
        g_layout.setMargin(0)
        g_layout.setSpacing(0)
        g_layout.addWidget(self.params)
        params_group.setLayout(g_layout)
        layout.addWidget(params_group)
        layout.setStretch(2,1000)

        self.cparams = QParamTable()
        params_group = QtGui.QGroupBox("Control Parameter Changes")
        g_layout = QtGui.QVBoxLayout()
        g_layout.setMargin(0)
        g_layout.setSpacing(0)
        g_layout.addWidget(self.cparams)
        params_group.setLayout(g_layout)
        layout.addWidget(params_group)
        layout.setStretch(3,1000)
        
        self.annotations = QParamTable()
        params_group = QtGui.QGroupBox("Annotation Changes")
        g_layout = QtGui.QVBoxLayout()
        g_layout.setMargin(0)
        g_layout.setSpacing(0)
        g_layout.addWidget(self.annotations)
        params_group.setLayout(g_layout)
        layout.addWidget(params_group)
        layout.setStretch(3,1000)
        
        self.setLayout(layout)
        self.addButtonsToToolbar()

    def addButtonsToToolbar(self):
        # Add the create analogy action
        self.createAnalogyAction = QtGui.QAction(
            CurrentTheme.VISUAL_DIFF_CREATE_ANALOGY_ICON,
            'Create analogy', None, triggered=self.createAnalogy)
        self.toolWindow().toolbar.insertAction(self.toolWindow().pinAction,
                                               self.createAnalogyAction)
        self.toolWindow().toolbar.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)

    def createAnalogy(self):
        if not hasattr(self.controller, 'current_diff'):
            return
        ((vistrail_a, version_a), (vistrail_b, version_b)) = \
            self.controller.current_diff_versions
        # analogy only works on single vistrail
        if not vistrail_a == vistrail_b:
            return


        default = 'from %s to %s' % (self.v1_name, self.v2_name)
        (result, ok) = QtGui.QInputDialog.getText(None, "Enter Analogy Name",
                                                  "Name of analogy:",
                                                  QtGui.QLineEdit.Normal,
                                                  default)
        if not ok:
            return
        result = str(result)
        try:
            self.controller.add_analogy(result, version_a, version_b)
        except VistrailsInternalError:
            debug.critical("Analogy name already exists")

    def set_diff(self):
        if not hasattr(self.controller, 'current_diff'):
            return
        ((vistrail_a, version_a), (vistrail_b, version_b)) = \
            self.controller.current_diff_versions

        # Set up the version name correctly
        v1_name = vistrail_a.getVersionName(version_a)
        if not v1_name:
            v1_name = 'Version %d' % version_a
        v2_name = vistrail_b.getVersionName(version_b)
        if not v2_name:
            v2_name = 'Version %d' % version_b

        # Add vistrail name if necessary
        if id(vistrail_a) != id(vistrail_b):
            if vistrail_a.locator is not None:
                v1_name = "%s : %s" % (vistrail_a.locator.short_name, v1_name)
            else:
                v1_name = "Vistrail A : %s" % v1_name
            if vistrail_b.locator is not None:
                v2_name = "%s : %s" % (vistrail_b.locator.short_name, v2_name)
            else:
                v2_name = "Vistrail B : %s" % v2_name

        self.v1_name, self.v2_name = v1_name, v2_name
        
        self.legend.set_names(v1_name, v2_name)
        self.params.set_names(v1_name, v2_name)
        self.cparams.set_names(v1_name, v2_name)
        self.annotations.set_names(v1_name, v2_name)
        self.update_module()

    def set_controller(self, controller=None):
        self.controller = controller
        if self.controller is not None:
            self.set_diff()
        
    def update_module(self, module=None):
        """ moduleSelected(id: int, selectedItems: [QGraphicsItem]) -> None
        When the user click on a module, display its parameter changes
        in the Inspector
        
        """
        if module is None or not hasattr(self.controller, 'current_diff'):
            self.params.model().clearList()
            self.params.parent().setVisible(False)
            self.cparams.model().clearList()
            self.cparams.parent().setVisible(False)
            self.annotations.model().clearList()
            self.annotations.parent().setVisible(False)
            return
        
        # Interpret the diff result and setup item models
        (p1, p2, v1Andv2, heuristicMatch, v1Only, v2Only, paramChanged,
         cparamChanged, annotChanged) = self.controller.current_diff

        # # Set the window title
        # if id>self.maxId1:
        #     self.inspector.setWindowTitle('Parameter Changes - %s' %
        #                                   p2.modules[id-self.maxId1-1].name)
        # else:
        #     self.inspector.setWindowTitle('Parameter Changes - %s' %
        #                                   p1.modules[id].name)

        # FIXME set the module name/package info?
            
        to_text = lambda x:'%s(%s)' % (x[0], ','.join(v[1] for v in x[1]))
        self.setTable(module, paramChanged, self.params, to_text)
        to_text = lambda x:'%s(%s)' % (x[0], x[1])
        self.setTable(module, cparamChanged, self.cparams, to_text)
        self.setTable(module, annotChanged, self.annotations, to_text)

    def setTable(self, module, changed, table, to_text):
        # Find the parameter changed module
        model = table.model()
        model.clearList()
        matching = None
        for ((m1id, m2id), paramMatching) in changed:
            if m1id == module.id:
                #print "found match"
                matching = paramMatching
                break

        #print "matching:", matching
        # If the module has no parameter changed, just display nothing
        if not matching:          
            table.parent().setVisible(False)
            return
        
        table.parent().setVisible(True)

        # Else just layout the diff on a table
        model.insertRows(0,len(matching))
        currentRow = 0
        for (f1, f2) in matching:
            if f1[0]!=None:
                model.setData(model.index(currentRow, 0), to_text(f1))
            if f2[0]!=None:
                model.setData(model.index(currentRow, 1), to_text(f2))
            if f1==f2:
                model.disableRow(currentRow)
            currentRow += 1

        table.resizeRowsToContents()

class QDiffView(QPipelineView):
    def __init__(self, parent=None):
        QPipelineView.__init__(self, parent)
        self.set_title("Diff")
        self.diff = None
        self.diff_versions = None
        self.controller = None
        self.setReadOnlyMode(True)
        
    def set_action_links(self):
        self.action_links = \
            {'copy': ('module_changed', self.has_selected_modules),
             'showGroup': ('module_changed', self.has_selected_group),
             'configureModule': ('module_changed', self.has_selected_module),
             'documentModule': ('module_changed', self.has_selected_module),
             }
            
    def set_action_defaults(self):
        self.action_defaults.update(
            {'execute'    : [('setEnabled', False, False)],
             'history'    : [('setEnabled', False, False)],
             'search'     : [('setEnabled', False, False)],
             'explore'    : [('setEnabled', False, False)],
             'provenance' : [('setEnabled', False, False)],
             'mashup'     : [('setEnabled', False, False)]
             })
            
    def set_to_current(self):
        self.controller.set_pipeline_view(self)
        self.controller.current_diff = self.diff
        self.controller.current_diff_versions = self.diff_versions

    def set_default_layout(self):
        from vistrails.gui.module_palette import QModulePalette
        self.set_palette_layout(
            {QtCore.Qt.LeftDockWidgetArea: QModulePalette,
             QtCore.Qt.RightDockWidgetArea: QDiffProperties,
             })
        
    def version_changed(self):
        pass

    def set_controller(self, controller):
        self.controller = controller
        self.scene().controller = controller

    def get_long_title(self):
        return self.long_title
    
    def set_diff_version_names(self):
        ((vistrail_a, version_a), (vistrail_b, version_b)) = self.diff_versions
        
        # Set up the version name correctly
        v1_name = vistrail_a.getVersionName(version_a)
        if not v1_name:
            v1_name = 'Version %d' % version_a
        v2_name = vistrail_b.getVersionName(version_b)
        if not v2_name:
            v2_name = 'Version %d' % version_b

        # Add vistrail name if necessary
        if id(vistrail_a) != id(vistrail_b):
            if vistrail_a.locator is not None:
                v1_name = "%s : %s" % (vistrail_a.locator.short_name, v1_name)
            else:
                v1_name = "Vistrail A : %s" % v1_name
            if vistrail_b.locator is not None:
                v2_name = "%s : %s" % (vistrail_b.locator.short_name, v2_name)
            else:
                v2_name = "Vistrail B : %s" % v2_name
            title = "Diff: %s x %s"%(v1_name, v2_name)
        else:
            title = "Diff: %s x %s from %s" % (v1_name, v2_name,
                                               self.vistrail_view.get_name())
        self.v1_name = v1_name
        self.v2_name = v2_name
        self.set_long_title(title)
        
    def set_diff(self, version_a, version_b, vistrail_b=None):
        # Interprete the diff result
        vistrail_a = self.controller.vistrail
        if vistrail_b is None:
            vistrail_b = self.controller.vistrail
        self.diff_versions = ((vistrail_a, version_a), 
                              (vistrail_b, version_b))
        self.set_diff_version_names()
        self.diff = vistrails.core.db.io.get_workflow_diff(*self.diff_versions)
            # self.controller.vistrail.get_pipeline_diff(version_a, version_b)
        (p1, p2, v1Andv2, heuristicMatch, v1Only, v2Only, paramChanged,
         cparamChanged, annotChanged) = self.diff
        # print "  $$$ v1Andv2:", v1Andv2
        # print "  $$$ heuristicMatch:", heuristicMatch
        # print "  $$$ v1Only", v1Only
        # print "  $$$ v2Only", v2Only
        # print "  $$$ paramChanged", paramChanged
        p1.validate(False)
        p2.validate(False)
        p_both = Pipeline()
        # the abstraction map is the same for both p1 and p2
        # p_both.set_abstraction_map(p1.abstraction_map)
        
        scene = self.scene()
        scene.clearItems()

        basic_pkg = get_vistrails_basic_pkg_id()

#         # FIXME HACK: We will create a dummy object that looks like a
#         # controller so that the qgraphicsmoduleitems and the scene
#         # are happy. It will simply store the pipeline will all
#         # modules and connections of the diff, and know how to copy stuff
#         class DummyController(object):
#             def __init__(self, pip):
#                 self.current_pipeline = pip
#                 self.search = None
#             def copy_modules_and_connections(self, module_ids, connection_ids):
#                 """copy_modules_and_connections(module_ids: [long],
#                                              connection_ids: [long]) -> str
#                 Serializes a list of modules and connections
#                 """

#                 pipeline = Pipeline()
# #                 pipeline.set_abstraction_map( \
# #                     self.current_pipeline.abstraction_map)
#                 for module_id in module_ids:
#                     module = self.current_pipeline.modules[module_id]
# #                     if module.vtType == Abstraction.vtType:
# #                         abstraction = \
# #                             pipeline.abstraction_map[module.abstraction_id]
# #                         pipeline.add_abstraction(abstraction)
#                     pipeline.add_module(module)
#                 for connection_id in connection_ids:
#                     connection = self.current_pipeline.connections[connection_id]
#                     pipeline.add_connection(connection)
#                 return core.db.io.serialize(pipeline)
                
#         controller = DummyController(p_both)
#         scene.controller = controller

        # Find the max version id from v1 and start the adding process
        self.maxId1 = 0
        for m1id in p1.modules.keys():
            if m1id>self.maxId1:
                self.maxId1 = m1id
        shiftId = self.maxId1 + 1

        # First add all shared modules, just use v1 module id
        sum1_x = 0.0
        sum1_y = 0.0
        sum2_x = 0.0
        sum2_y = 0.0
        for (m1id, m2id) in chain(v1Andv2, heuristicMatch):
        #     item = scene.addModule(p1.modules[m1id],
        #                            CurrentTheme.VISUAL_DIFF_SHARED_BRUSH)
        #     item.controller = self.controller
        #     p_both.add_module(copy.copy(p1.modules[m1id]))
        #     sum1_x += p1.modules[m1id].location.x
        #     sum1_y += p1.modules[m1id].location.y
        #     sum2_x += p2.modules[m2id].location.x
        #     sum2_y += p2.modules[m2id].location.y
        # for (m1id, m2id) in heuristicMatch:
            m1 = p1.modules[m1id]
            m2 = p2.modules[m2id]

            sum1_x += p1.modules[m1id].location.x
            sum1_y += p1.modules[m1id].location.y
            sum2_x += p2.modules[m2id].location.x
            sum2_y += p2.modules[m2id].location.y            

            #this is a hack for modules with a dynamic local registry.
            #The problem arises when modules have the same name but different
            #input/output ports. We just make sure that the module we add to
            # the canvas has the ports from both modules, so we don't have
            # addconnection errors.
            port_specs = dict(((p.type, p.name), p) for p in m1.port_spec_list)
            for p in m2.port_spec_list:
                p_key = (p.type, p.name)
                if not p_key in port_specs:
                    m1.add_port_spec(p)
                elif port_specs[p_key] != p:
                    #if we add this port, we will get port overloading.
                    #To avoid this, just cast the current port to be of
                    # Module or Variant type.
                    old_port_spec = port_specs[p_key]
                    m1.delete_port_spec(old_port_spec)
                    if old_port_spec.type == 'input':
                        m_sig = '(%s:Module)' % basic_pkg
                    else:
                        m_sig = '(%s:Variant)' % basic_pkg
                    new_port_spec = PortSpec(id=old_port_spec.id,
                                             name=old_port_spec.name,
                                             type=old_port_spec.type,
                                             sigstring=m_sig)
                    m1.add_port_spec(new_port_spec)

            item = scene.addModule(p1.modules[m1id],
                                   CurrentTheme.VISUAL_DIFF_SHARED_BRUSH)
            # item = scene.addModule(p1.modules[m1id],
            #                        CurrentTheme.VISUAL_DIFF_MATCH_BRUSH)
            item.controller = self.controller
            p_both.add_module(copy.copy(p1.modules[m1id]))

        # Then add parameter changed version
        inChanged = set(m for (m, matching)
                        in chain(paramChanged, cparamChanged, annotChanged))
        for (m1id, m2id) in inChanged:
            m1 = p1.modules[m1id]
            m2 = p2.modules[m2id]
            sum1_x += p1.modules[m1id].location.x
            sum1_y += p1.modules[m1id].location.y
            sum2_x += p2.modules[m2id].location.x
            sum2_y += p2.modules[m2id].location.y

            #this is a hack for modules with a dynamic local registry.
            #The problem arises when modules have the same name but different
            #input/output ports. We just make sure that the module we add to
            # the canvas has the ports from both modules, so we don't have
            # addconnection errors.
            port_specs = dict(((p.type, p.name), p) for p in m1.port_spec_list)
            for p in m2.port_spec_list:
                p_key = (p.type, p.name)
                if not p_key in port_specs:
                    m1.add_port_spec(p)
                elif port_specs[p_key] != p:
                    #if we add this port, we will get port overloading.
                    #To avoid this, just cast the current port to be of
                    # Module or Variant type.
                    old_port_spec = port_specs[p_key]
                    m1.delete_port_spec(old_port_spec)
                    if old_port_spec.type == 'input':
                        m_sig = '(%s:Module)' % basic_pkg
                    else:
                        m_sig = '(%s:Variant)' % basic_pkg
                    new_port_spec = PortSpec(id=old_port_spec.id,
                                             name=old_port_spec.name,
                                             type=old_port_spec.type,
                                             sigstring=m_sig)
                    m1.add_port_spec(new_port_spec)
                            
            item = scene.addModule(p1.modules[m1id],
                                   CurrentTheme.VISUAL_DIFF_PARAMETER_CHANGED_BRUSH)
            item.controller = self.controller
            p_both.add_module(copy.copy(p1.modules[m1id]))

        total_len = len(v1Andv2) + + len(heuristicMatch) + len(paramChanged)
        if total_len != 0:
            avg1_x = sum1_x / total_len
            avg1_y = sum1_y / total_len
            avg2_x = sum2_x / total_len
            avg2_y = sum2_y / total_len
        else:
            avg1_x = 0.0
            avg1_y = 0.0
            avg2_x = 0.0
            avg2_y = 0.0
#         avg1_x = sum1_x / total_len if total_len != 0 else 0.0
#         avg1_y = sum1_y / total_len if total_len != 0 else 0.0
#         avg2_x = sum2_x / total_len if total_len != 0 else 0.0
#         avg2_y = sum2_y / total_len if total_len != 0 else 0.0

        # Now add the ones only applicable for v1, still using v1 ids
        for m1id in v1Only:
            item = scene.addModule(p1.modules[m1id],
                                   CurrentTheme.VISUAL_DIFF_FROM_VERSION_BRUSH)
            item.controller = self.controller
            p_both.add_module(copy.copy(p1.modules[m1id]))

        # Now add the ones for v2 only but must shift the ids away from v1
        for m2id in v2Only:
            p2_module = copy.copy(p2.modules[m2id])
            p2_module.id = m2id + shiftId
            # p2.modules[m2id].id = m2id + shiftId
            p2_module.location.x -= avg2_x - avg1_x
            p2_module.location.y -= avg2_y - avg1_y
            item = scene.addModule(p2_module, #p2.modules[m2id],
                                   CurrentTheme.VISUAL_DIFF_TO_VERSION_BRUSH)
            item.controller = self.controller
            # p_both.add_module(copy.copy(p2.modules[m2id]))
            p_both.add_module(p2_module)
            

        # Create a mapping between share modules between v1 and v2
        v1Tov2 = {}
        v2Tov1 = {}
        for (m1id, m2id) in v1Andv2:
            v1Tov2[m1id] = m2id
            v2Tov1[m2id] = m1id
        for (m1id, m2id) in heuristicMatch:
            v1Tov2[m1id] = m2id
            v2Tov1[m2id] = m1id
        for (m1id, m2id) in inChanged:
            v1Tov2[m1id] = m2id
            v2Tov1[m2id] = m1id

        # Next we're going to add connections, only connections of
        # v2Only need to shift their ids
        if p1.connections.keys():
            connectionShift = max(p1.connections.keys())+1
        else:
            connectionShift = 0
        allConnections = copy.copy(p1.connections)
        sharedConnections = []
        v2OnlyConnections = []        
        for (cid2, connection2) in copy.copy(p2.connections.items()):
            if connection2.sourceId in v2Only:
                connection2.sourceId += shiftId
            else:
                connection2.sourceId = v2Tov1[connection2.sourceId]
                
            if connection2.destinationId in v2Only:
                connection2.destinationId += shiftId
            else:
                connection2.destinationId = v2Tov1[connection2.destinationId]

            # Is this connection also shared by p1?
            shared = False
            for (cid1, connection1) in p1.connections.items():
                if (connection1.sourceId==connection2.sourceId and
                    connection1.destinationId==connection2.destinationId and
                    connection1.source.name==connection2.source.name and
                    connection1.destination.name==connection2.destination.name):
                    sharedConnections.append(cid1)
                    shared = True
                    break
            if not shared:
                allConnections[cid2+connectionShift] = connection2
                connection2.id = cid2+connectionShift
                v2OnlyConnections.append(cid2+connectionShift)

        connectionItems = []
        for c in allConnections.values():
            p_both.add_connection(copy.copy(c))
            connectionItems.append(scene.addConnection(c))

        # Color Code connections
        for c in connectionItems:
            if c.id in sharedConnections:
                pass
            elif c.id in v2OnlyConnections:
                pen = QtGui.QPen(CurrentTheme.CONNECTION_PEN)
                pen.setBrush(CurrentTheme.VISUAL_DIFF_TO_VERSION_BRUSH)
                c.connectionPen = pen
            else:
                pen = QtGui.QPen(CurrentTheme.CONNECTION_PEN)
                pen.setBrush(CurrentTheme.VISUAL_DIFF_FROM_VERSION_BRUSH)
                c.connectionPen = pen

       
        scene.current_pipeline = p_both

        scene.updateSceneBoundingRect()
        scene.fitToView(self, True)

class QVisualDiff(QtGui.QMainWindow):
    """
    QVisualDiff is a main widget for Visual Diff containing a GL
    Widget to draw the pipeline
    
    """
    def __init__(self, vistrail, v1, v2,
                 controller,
                 parent=None, f=QtCore.Qt.WindowFlags()):
        """ QVisualDiff(vistrail: Vistrail, v1: str, v2: str,
                        parent: QWidget, f: WindowFlags) -> QVisualDiff
        Initialize with all
        
        """
        # Set up the version name correctly
        v1Name = vistrail.getVersionName(v1)
        if not v1Name: v1Name = 'Version %d' % v1
        v2Name = vistrail.getVersionName(v2)        
        if not v2Name: v2Name = 'Version %d' % v2
        
        # Actually perform the diff and store its result
        self.diff = vistrail.get_pipeline_diff(v1, v2)

        self.v1_name = v1Name
        self.v2_name = v2Name
        self.v1 = v1
        self.v2 = v2
        self.controller = controller

        # Create the top-level Visual Diff window
        windowDecors = f | QtCore.Qt.Dialog |QtCore.Qt.WindowMaximizeButtonHint
        QtGui.QMainWindow.__init__(self, parent)
        self.setWindowTitle('Visual Diff - from %s to %s' % (v1Name, v2Name))
        self.setMouseTracking(True)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setSizePolicy(QtGui.QSizePolicy(QtGui.QSizePolicy.Expanding,
                                             QtGui.QSizePolicy.Expanding))
        self.createPipelineView()
        self.createToolBar()
        self.createToolWindows(v1Name, v2Name)

        self.installEventFilter(self)

    def eventFilter(self, object, event):
        if event.type() == QtCore.QEvent.ShortcutOverride and \
                event.key() == QtCore.Qt.Key_W and \
                event.modifiers() == QtCore.Qt.ControlModifier:
            event.accept()
            self.close()
        return QtGui.QMainWindow.eventFilter(self, object, event)

    def closeEvent(self, event):
        self.inspector.close()
        self.legendWindow.close()

    def createPipelineView(self):
        """ createPipelineView() -> None        
        Create a center pipeline view for Visual Diff and setup the
        view based on the diff result self.diff
        
        """
        # Initialize the shape engine
        self.pipelineView = QPipelineView()
        self.setCentralWidget(self.pipelineView)

        # Add all the shapes into the view
        self.createDiffPipeline()

        # Hook shape selecting functions
        self.connect(self.pipelineView.scene(), QtCore.SIGNAL("moduleSelected"),
                     self.moduleSelected)

    def createToolBar(self):
        """ createToolBar() -> None        
        Create the default toolbar of Visual Diff window with two
        buttons to toggle the Parameter Inspector and Legend window
        
        """
        # Initialize the Visual Diff toolbar
        self.toolBar = self.addToolBar('Visual Diff Toolbar')
        self.toolBar.setMovable(False)

        # Add the Show Parameter Inspector action
        self.showInspectorAction = self.toolBar.addAction(
            CurrentTheme.VISUAL_DIFF_SHOW_PARAM_ICON,
            'Show Parameter Inspector window')
        self.showInspectorAction.setCheckable(True)
        self.connect(self.showInspectorAction, QtCore.SIGNAL("toggled(bool)"),
                     self.toggleShowInspector)
        
        # Add the Show Legend window action
        self.showLegendsAction = self.toolBar.addAction(
            CurrentTheme.VISUAL_DIFF_SHOW_LEGEND_ICON,
            'Show Legends')
        self.showLegendsAction.setCheckable(True)
        self.connect(self.showLegendsAction, QtCore.SIGNAL("toggled(bool)"),
                     self.toggleShowLegend)

        # Add the create analogy action
        self.createAnalogyAction = self.toolBar.addAction(
            CurrentTheme.VISUAL_DIFF_CREATE_ANALOGY_ICON,
            'Create analogy')
        self.connect(self.createAnalogyAction, QtCore.SIGNAL("triggered()"),
                     self.createAnalogy)

    def createAnalogy(self):
        default = 'from %s to %s' % (self.v1_name, self.v2_name)
        (result, ok) = QtGui.QInputDialog.getText(None, "Enter Analogy Name",
                                                  "Name of analogy:",
                                                  QtGui.QLineEdit.Normal,
                                                  default)
        if not ok:
            return
        result = str(result)
        try:
            self.controller.add_analogy(result, self.v1, self.v2)
        except VistrailsInternalError:
            debug.critical("Analogy name already exists")
        
    def createToolWindows(self, v1Name, v2Name):
        """ createToolWindows(v1Name: str, v2Name: str) -> None
        Create Inspector and Legend window

        """
        self.inspector = QParamInspector(v1Name, v2Name, self)
        self.inspector.resize(QtCore.QSize(
            *CurrentTheme.VISUAL_DIFF_PARAMETER_WINDOW_SIZE))
        self.legendWindow = QLegendWindow(v1Name, v2Name,self)

    def moduleSelected(self, id, selectedItems):
        """ moduleSelected(id: int, selectedItems: [QGraphicsItem]) -> None
        When the user click on a module, display its parameter changes
        in the Inspector
        
        """
        if len(selectedItems)!=1 or id==-1:
            self.moduleUnselected()
            return
        
        # Interprete the diff result and setup item models
        (p1, p2, v1Andv2, heuristicMatch, v1Only, v2Only, paramChanged) = \
            self.diff

        # Set the window title
        if id>self.maxId1:
            self.inspector.setWindowTitle('Parameter Changes - %s' %
                                          p2.modules[id-self.maxId1-1].name)
        else:
            self.inspector.setWindowTitle('Parameter Changes - %s' %
                                          p1.modules[id].name)
            
        # Clear the old inspector
        functions = self.inspector.functionsTab.model()
#         annotations = self.inspector.annotationsTab.model()
        functions.clearList()
#         annotations.clearList()

        # Find the parameter changed module
        matching = None
        for ((m1id, m2id), paramMatching) in paramChanged:
            if m1id==id:
                matching = paramMatching
                break

        # If the module has no parameter changed, just display nothing
        if not matching:          
            return
        
        # Else just layout the diff on a table
        functions.insertRows(0,len(matching))
        currentRow = 0
        for (f1, f2) in matching:
            if f1[0]!=None:
                functions.setData(
                    functions.index(currentRow, 0),
                    '%s(%s)' % (f1[0], ','.join(v[1] for v in f1[1])))
            if f2[0]!=None:
                functions.setData(
                    functions.index(currentRow, 1),
                    '%s(%s)' % (f2[0], ','.join(v[1] for v in f2[1])))
            if f1==f2:                
                functions.disableRow(currentRow)
            currentRow += 1

        self.inspector.functionsTab.resizeRowsToContents()
#         self.inspector.annotationsTab.resizeRowsToContents()

    def moduleUnselected(self):
        """ moduleUnselected() -> None
        When a user selects nothing, make sure to display nothing as well
        
        """
#         self.inspector.annotationsTab.model().clearList()
        self.inspector.functionsTab.model().clearList()
        self.inspector.setWindowTitle('Parameter Changes - None')

    def toggleShowInspector(self):
        """ toggleShowInspector() -> None
        Show/Hide the inspector when toggle this button
        
        """
        if self.inspector.firstTime:
            max_geom = QtGui.QApplication.desktop().screenGeometry(self)
            if (self.frameSize().width() <
                max_geom.width() - self.inspector.width()):
                self.inspector.move(self.pos().x()+self.frameSize().width(),
                                    self.pos().y())
            else:
                self.inspector.move(self.pos().x()+self.frameSize().width()-
                                   self.inspector.frameSize().width(),
                                   self.pos().y() +
                                    self.legendWindow.frameSize().height())
            self.inspector.firstTime = False
        self.inspector.setVisible(self.showInspectorAction.isChecked())
            
    def toggleShowLegend(self):
        """ toggleShowLegend() -> None
        Show/Hide the legend window when toggle this button
        
        """
        if self.legendWindow.firstTime:
            self.legendWindow.move(self.pos().x()+self.frameSize().width()-
                                   self.legendWindow.frameSize().width(),
                                   self.pos().y())
        self.legendWindow.setVisible(self.showLegendsAction.isChecked())
        if self.legendWindow.firstTime:
            self.legendWindow.firstTime = False
            self.legendWindow.setFixedSize(self.legendWindow.size())            
                
    def createDiffPipeline(self):
        """ createDiffPipeline() -> None        
        Actually walk through the self.diff result and add all modules
        into the pipeline view
        
        """

        # Interprete the diff result
        (p1, p2, v1Andv2, heuristicMatch, v1Only, v2Only, paramChanged) = \
            self.diff
        self.controller.validate(p1, False)
        self.controller.validate(p2, False)
        p_both = Pipeline()
        # the abstraction map is the same for both p1 and p2
        # p_both.set_abstraction_map(p1.abstraction_map)
        
        scene = self.pipelineView.scene()
        scene.clearItems()

        basic_pkg = get_vistrails_basic_pkg_id()

        # FIXME HACK: We will create a dummy object that looks like a
        # controller so that the qgraphicsmoduleitems and the scene
        # are happy. It will simply store the pipeline will all
        # modules and connections of the diff, and know how to copy stuff
        class DummyController(object):
            def __init__(self, pip):
                self.current_pipeline = pip
                self.search = None
            def copy_modules_and_connections(self, module_ids, connection_ids):
                """copy_modules_and_connections(module_ids: [long],
                                             connection_ids: [long]) -> str
                Serializes a list of modules and connections
                """

                pipeline = Pipeline()
#                 pipeline.set_abstraction_map( \
#                     self.current_pipeline.abstraction_map)
                for module_id in module_ids:
                    module = self.current_pipeline.modules[module_id]
#                     if module.vtType == Abstraction.vtType:
#                         abstraction = \
#                             pipeline.abstraction_map[module.abstraction_id]
#                         pipeline.add_abstraction(abstraction)
                    pipeline.add_module(module)
                for connection_id in connection_ids:
                    connection = self.current_pipeline.connections[connection_id]
                    pipeline.add_connection(connection)
                return vistrails.core.db.io.serialize(pipeline)
                
        controller = DummyController(p_both)
        scene.controller = controller

        # Find the max version id from v1 and start the adding process
        self.maxId1 = 0
        for m1id in p1.modules.keys():
            if m1id>self.maxId1:
                self.maxId1 = m1id
        shiftId = self.maxId1 + 1

        # First add all shared modules, just use v1 module id
        sum1_x = 0.0
        sum1_y = 0.0
        sum2_x = 0.0
        sum2_y = 0.0
        for (m1id, m2id) in v1Andv2:
            item = scene.addModule(p1.modules[m1id],
                                   CurrentTheme.VISUAL_DIFF_SHARED_BRUSH)
            item.controller = controller
            p_both.add_module(copy.copy(p1.modules[m1id]))
            sum1_x += p1.modules[m1id].location.x
            sum1_y += p1.modules[m1id].location.y
            sum2_x += p2.modules[m2id].location.x
            sum2_y += p2.modules[m2id].location.y
        for (m1id, m2id) in heuristicMatch:
            m1 = p1.modules[m1id]
            m2 = p2.modules[m2id]

            sum1_x += p1.modules[m1id].location.x
            sum1_y += p1.modules[m1id].location.y
            sum2_x += p2.modules[m2id].location.x
            sum2_y += p2.modules[m2id].location.y            

            #this is a hack for modules with a dynamic local registry.
            #The problem arises when modules have the same name but different
            #input/output ports. We just make sure that the module we add to
            # the canvas has the ports from both modules, so we don't have
            # addconnection errors.
            port_specs = dict(((p.type, p.name), p) for p in m1.port_spec_list)
            for p in m2.port_spec_list:
                p_key = (p.type, p.name)
                if not p_key in port_specs:
                    m1.add_port_spec(p)
                elif port_specs[p_key] != p:
                    #if we add this port, we will get port overloading.
                    #To avoid this, just cast the current port to be of
                    # Module or Variant type.
                    old_port_spec = port_specs[p_key]
                    m1.delete_port_spec(old_port_spec)
                    if old_port_spec.type == 'input':
                        m_sig = '(%s:Module)' % basic_pkg
                    else:
                        m_sig = '(%s:Variant)' % basic_pkg
                    new_port_spec = PortSpec(id=old_port_spec.id,
                                             name=old_port_spec.name,
                                             type=old_port_spec.type,
                                             sigstring=m_sig)
                    m1.add_port_spec(new_port_spec)

            item = scene.addModule(p1.modules[m1id],
                                   CurrentTheme.VISUAL_DIFF_MATCH_BRUSH)
            item.controller = controller
            p_both.add_module(copy.copy(p1.modules[m1id]))

        # Then add parameter changed version
        for ((m1id, m2id), matching) in paramChanged:
            m1 = p1.modules[m1id]
            m2 = p2.modules[m2id]
            
            sum1_x += p1.modules[m1id].location.x
            sum1_y += p1.modules[m1id].location.y
            sum2_x += p2.modules[m2id].location.x
            sum2_y += p2.modules[m2id].location.y

            #this is a hack for modules with a dynamic local registry.
            #The problem arises when modules have the same name but different
            #input/output ports. We just make sure that the module we add to
            # the canvas has the ports from both modules, so we don't have
            # addconnection errors.
            port_specs = dict(((p.type, p.name), p) for p in m1.port_spec_list)
            for p in m2.port_spec_list:
                p_key = (p.type, p.name)
                if not p_key in port_specs:
                    m1.add_port_spec(p)
                elif port_specs[p_key] != p:
                    #if we add this port, we will get port overloading.
                    #To avoid this, just cast the current port to be of
                    # Module or Variant type.
                    old_port_spec = port_specs[p_key]
                    m1.delete_port_spec(old_port_spec)
                    if old_port_spec.type == 'input':
                        m_sig = '(%s:Module)' % basic_pkg
                    else:
                        m_sig = '(%s:Variant)' % basic_pkg
                    new_port_spec = PortSpec(id=old_port_spec.id,
                                             name=old_port_spec.name,
                                             type=old_port_spec.type,
                                             sigstring=m_sig)
                    m1.add_port_spec(new_port_spec)

            item = scene.addModule(p1.modules[m1id],
                                   CurrentTheme.VISUAL_DIFF_PARAMETER_CHANGED_BRUSH)
            item.controller = controller
            p_both.add_module(copy.copy(p1.modules[m1id]))

        total_len = len(v1Andv2) + + len(heuristicMatch) + len(paramChanged)
        if total_len != 0:
            avg1_x = sum1_x / total_len
            avg1_y = sum1_y / total_len
            avg2_x = sum2_x / total_len
            avg2_y = sum2_y / total_len
        else:
            avg1_x = 0.0
            avg1_y = 0.0
            avg2_x = 0.0
            avg2_y = 0.0
#         avg1_x = sum1_x / total_len if total_len != 0 else 0.0
#         avg1_y = sum1_y / total_len if total_len != 0 else 0.0
#         avg2_x = sum2_x / total_len if total_len != 0 else 0.0
#         avg2_y = sum2_y / total_len if total_len != 0 else 0.0

        # Now add the ones only applicable for v1, still using v1 ids
        for m1id in v1Only:
            item = scene.addModule(p1.modules[m1id],
                                   CurrentTheme.VISUAL_DIFF_FROM_VERSION_BRUSH)
            item.controller = controller
            p_both.add_module(copy.copy(p1.modules[m1id]))

        # Now add the ones for v2 only but must shift the ids away from v1
        for m2id in v2Only:
            p2_module = copy.copy(p2.modules[m2id])
            p2_module.id = m2id + shiftId
            # p2.modules[m2id].id = m2id + shiftId
            p2_module.location.x -= avg2_x - avg1_x
            p2_module.location.y -= avg2_y - avg1_y
            item = scene.addModule(p2_module, #p2.modules[m2id],
                                   CurrentTheme.VISUAL_DIFF_TO_VERSION_BRUSH)
            item.controller = controller
            # p_both.add_module(copy.copy(p2.modules[m2id]))
            p_both.add_module(p2_module)
            

        # Create a mapping between share modules between v1 and v2
        v1Tov2 = {}
        v2Tov1 = {}
        for (m1id, m2id) in v1Andv2:
            v1Tov2[m1id] = m2id
            v2Tov1[m2id] = m1id
        for (m1id, m2id) in heuristicMatch:
            v1Tov2[m1id] = m2id
            v2Tov1[m2id] = m1id
        for ((m1id, m2id), matching) in paramChanged:
            v1Tov2[m1id] = m2id
            v2Tov1[m2id] = m1id

        # Next we're going to add connections, only connections of
        # v2Only need to shift their ids
        if p1.connections.keys():
            connectionShift = max(p1.connections.keys())+1
        else:
            connectionShift = 0
        allConnections = copy.copy(p1.connections)
        sharedConnections = []
        v2OnlyConnections = []        
        for (cid2, connection2) in copy.copy(p2.connections.items()):
            if connection2.sourceId in v2Only:
                connection2.sourceId += shiftId
            else:
                connection2.sourceId = v2Tov1[connection2.sourceId]
                
            if connection2.destinationId in v2Only:
                connection2.destinationId += shiftId
            else:
                connection2.destinationId = v2Tov1[connection2.destinationId]

            # Is this connection also shared by p1?
            shared = False
            for (cid1, connection1) in p1.connections.items():
                if (connection1.sourceId==connection2.sourceId and
                    connection1.destinationId==connection2.destinationId and
                    connection1.source.name==connection2.source.name and
                    connection1.destination.name==connection2.destination.name):
                    sharedConnections.append(cid1)
                    shared = True
                    break
            if not shared:
                allConnections[cid2+connectionShift] = connection2
                connection2.id = cid2+connectionShift
                v2OnlyConnections.append(cid2+connectionShift)

        connectionItems = []
        for c in allConnections.values():
            p_both.add_connection(copy.copy(c))
            connectionItems.append(scene.addConnection(c))

        # Color Code connections
        for c in connectionItems:
            if c.id in sharedConnections:
                pass
            elif c.id in v2OnlyConnections:
                pen = QtGui.QPen(CurrentTheme.CONNECTION_PEN)
                pen.setBrush(CurrentTheme.VISUAL_DIFF_TO_VERSION_BRUSH)
                c.connectionPen = pen
            else:
                pen = QtGui.QPen(CurrentTheme.CONNECTION_PEN)
                pen.setBrush(CurrentTheme.VISUAL_DIFF_FROM_VERSION_BRUSH)
                c.connectionPen = pen

       

        scene.updateSceneBoundingRect()
        scene.fitToView(self.pipelineView, True)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

import base64
import copy
import gc
import cPickle as pickle

from vistrails.core.common import InstanceObject, VistrailsInternalError
from vistrails.core.data_structures.bijectivedict import Bidict
from vistrails.core import debug
import vistrails.core.interpreter.base
from vistrails.core.interpreter.base import AbortExecution
import vistrails.core.interpreter.utils
from vistrails.core.log.controller import DummyLogController
from vistrails.core.modules.basic_modules import identifier as basic_pkg, \
                                                 Generator
from vistrails.core.modules.module_registry import get_module_registry
from vistrails.core.modules.vistrails_module import ModuleBreakpoint, \
    ModuleConnector, ModuleError, ModuleErrors, ModuleHadError, \
    ModuleSuspended, ModuleWasSuspended
from vistrails.core.utils import DummyView
import vistrails.core.system
import vistrails.core.vistrail.pipeline


###############################################################################

class ViewUpdatingLogController(object):
    class Loop(object):
        def __init__(self, logger, view):
            self.log = logger
            self.view = view

        def end_loop_execution(self):
            self.log.finish_loop_execution()

        def begin_iteration(self, looped_obj, iteration):
            self.log.start_iteration(looped_obj, iteration)

        def end_iteration(self, looped_obj):
            self.log.finish_iteration(looped_obj)

    def __init__(self, logger, view, remap_id, ids,
                 module_executed_hook=[]):
        self.log = logger
        self.view = view
        self.remap_id = remap_id
        self.ids = set(ids) # modules left to be executed
        self.nb_modules = len(self.ids)
        self.module_executed_hook = module_executed_hook

        self.errors = {}
        self.executed = {}
        self.suspended = {}
        self.cached = {}

    def signalSuccess(self, obj):
        self.executed[obj.id] = True
        for callable_ in self.module_executed_hook:
            callable_(obj.id)

    def signalError(self, obj, error):
        self.errors[obj.id] = error

    def begin_update(self, obj):
        i = self.remap_id(obj.id)
        self.view.set_module_active(i)

    def begin_compute(self, obj):
        i = self.remap_id(obj.id)
        self.view.set_module_computing(i)

        reg = get_module_registry()
        module_name = reg.get_descriptor(obj.__class__).name

        self.log.start_execution(obj, i, module_name)

    def update_progress(self, obj, progress=0.0):
        i = self.remap_id(obj.id)
        self.view.set_module_progress(i, progress)

    def begin_loop_execution(self, obj, total_iterations=None):
        return ViewUpdatingLogController.Loop(
                self.log.start_loop_execution(obj, total_iterations),
                self.view)

    def _handle_suspended(self, obj, error):
        """ _handle_suspended(obj: VistrailsModule, error: ModuleSuspended
            ) -> None
            Report module as suspended
        """
        # update job monitor because this may be an old-style job
        jm = obj.job_monitor()
        reg = get_module_registry()
        name = reg.get_descriptor(obj.__class__).name
        iteration = self.log.get_iteration_from_module(obj)
        if iteration is not None:
            name += '/%d' % iteration
        # add to parent list for computing the module tree later
        error.name = name
        jm.addParent(error)

    def end_update(self, obj, error=None, errorTrace=None,
            was_suspended=False):
        try:
            i = self.remap_id(obj.id)
        except KeyError:
            # This happens with Groups: we get end_update for modules inside
            # the Group, which we can't remap to the current pipeline
            # It's ok, because that was already logged by the recursive
            # execute_pipeline() call
            return
        if was_suspended:
            self._handle_suspended(obj, error)
            self.suspended[obj.id] = error
            self.view.set_module_suspended(i, error)
            if error.children:
                for child in error.children:
                    self.end_update(child.module, child, was_suspended=True)
        elif error is None:
            self.view.set_module_success(i)
        else:
            self.view.set_module_error(i, error)

        if i in self.ids:
            self.ids.remove(i)
            self.view.set_execution_progress(
                    1.0 - ((len(self.ids) + len(Generator.generators)) * 1.0 /
                           (self.nb_modules + len(Generator.generators))))

        msg = '' if error is None else error.msg
        self.log.finish_execution(obj, msg, errorTrace,
                                  was_suspended)

    def update_cached(self, obj):
        self.cached[obj.id] = True
        i = self.remap_id(obj.id)

        reg = get_module_registry()
        module_name = reg.get_descriptor(obj.__class__).name

        self.log.start_execution(obj, i, module_name,
                                 cached=1)
        self.view.set_module_not_executed(i)
        self.log.finish_execution(obj, '')

    def set_computing(self, obj):
        i = self.remap_id(obj.id)
        self.view.set_module_computing(i)

    def annotate(self, obj, d):
        self.log.insert_module_annotations(obj, d)

    def add_machine(self, machine):
        return self.log.add_machine(machine)

    def add_exec(self, exec_):
        return self.log.add_exec(exec_)

###############################################################################

Variant_desc = None
InputPort_desc = None

class CachedInterpreter(vistrails.core.interpreter.base.BaseInterpreter):

    def __init__(self):
        vistrails.core.interpreter.base.BaseInterpreter.__init__(self)
        self.debugger = None
        self.create()

    def create(self):
        # FIXME moved here because otherwise we hit the registry too early
        from vistrails.core.modules.module_utils import FilePool
        self._file_pool = FilePool()
        self._persistent_pipeline = vistrails.core.vistrail.pipeline.Pipeline()
        self._objects = {}
        self.filePool = self._file_pool
        self._streams = []

    def clear(self):
        self._file_pool.cleanup()
        self._persistent_pipeline.clear()
        for obj in self._objects.itervalues():
            obj.clear()
        self._objects = {}

    def __del__(self):
        self.clear()

    def clean_modules(self, modules_to_clean):
        """clean_modules(modules_to_clean: list of persistent module ids)

        Removes modules from the persistent pipeline, and the modules that
        depend on them."""
        if not modules_to_clean:
            return
        g = self._persistent_pipeline.graph
        modules_to_clean = (set(modules_to_clean) &
                            set(self._persistent_pipeline.modules.iterkeys()))
        dependencies = g.vertices_topological_sort(modules_to_clean)
        for v in dependencies:
            self._persistent_pipeline.delete_module(v)
            del self._objects[v]

    def clean_non_cacheable_modules(self):
        """clean_non_cacheable_modules() -> None

        Removes all modules that are not cacheable from the persistent
        pipeline, and the modules that depend on them.
        """
        non_cacheable_modules = [i for
                                 (i, mod) in self._objects.iteritems()
                                 if not mod.is_cacheable()]
        self.clean_modules(non_cacheable_modules)

    def _clear_package(self, identifier):
        """clear_package(identifier: str) -> None

        Removes all modules from the given package from the persistent
        pipeline.
        """
        modules = [mod.id
                   for mod in self._persistent_pipeline.module_list
                   if mod.module_descriptor.identifier == identifier]
        self.clean_modules(modules)

    def make_connection(self, conn, src, dst):
        """make_connection(self, conn, src, dst)
        Builds a execution-time connection between modules.

        """
        iport = conn.destination.name
        oport = conn.source.name
        src.enable_output_port(oport)
        src.load_type_check_descs()
        if isinstance(src, src.InputPort_desc.module):
            typecheck = [False]
        else:
            typecheck = src.get_type_checks(conn.source.spec)
        dst.set_input_port(iport,
                           ModuleConnector(src, oport, conn.source.spec,
                                           typecheck))

    def setup_pipeline(self, pipeline, **kwargs):
        """setup_pipeline(controller, pipeline, locator, currentVersion,
                          view, aliases, **kwargs)
        Matches a pipeline with the persistent pipeline and creates
        instances of modules that aren't in the cache.
        """
        def fetch(name, default):
            return kwargs.pop(name, default)
        controller = fetch('controller', None)
        locator = fetch('locator', None)
        current_version = fetch('current_version', None)
        view = fetch('view', DummyView())
        vistrail_variables = fetch('vistrail_variables', None)
        aliases = fetch('aliases', None)
        params = fetch('params', None)
        extra_info = fetch('extra_info', None)
        logger = fetch('logger', DummyLogController)
        sinks = fetch('sinks', None)
        reason = fetch('reason', None)
        actions = fetch('actions', None)
        done_summon_hooks = fetch('done_summon_hooks', [])
        module_executed_hook = fetch('module_executed_hook', [])
        stop_on_error = fetch('stop_on_error', True)
        parent_exec = fetch('parent_exec', None)

        reg = get_module_registry()

        if len(kwargs) > 0:
            raise VistrailsInternalError('Wrong parameters passed '
                                         'to setup_pipeline: %s' % kwargs)

        def create_null():
            """Creates a Null value"""
            getter = reg.get_descriptor_by_name
            descriptor = getter(basic_pkg, 'Null')
            return descriptor.module()
        
        def create_constant(param, module):
            """Creates a Constant from a parameter spec"""
            getter = reg.get_descriptor_by_name
            desc = getter(param.identifier, param.type, param.namespace)
            constant = desc.module()
            constant.id = module.id
#             if param.evaluatedStrValue:
#                 constant.setValue(param.evaluatedStrValue)
            if param.strValue != '':
                constant.setValue(param.strValue)
            else:
                constant.setValue( \
                    constant.translate_to_string(constant.default_value))
            return constant

        ### BEGIN METHOD ###

#         if self.debugger:
#             self.debugger.update()
        to_delete = []
        errors = {}

        if controller is not None:
            # Controller is none for sub_modules
            controller.validate(pipeline)
        else:
            pipeline.validate()

        self.resolve_aliases(pipeline, aliases)
        if vistrail_variables:
            self.resolve_variables(vistrail_variables,  pipeline)

        self.update_params(pipeline, params)
        
        (tmp_to_persistent_module_map,
         conn_map,
         module_added_set,
         conn_added_set) = self.add_to_persistent_pipeline(pipeline)

        # Create the new objects
        for i in module_added_set:
            persistent_id = tmp_to_persistent_module_map[i]
            module = self._persistent_pipeline.modules[persistent_id]
            obj = self._objects[persistent_id] = module.summon()
            obj.interpreter = self
            obj.id = persistent_id
            obj.signature = module._signature
            
            # Checking if output should be stored
            if module.has_annotation_with_key('annotate_output'):
                annotate_output = module.get_annotation_by_key('annotate_output')
                #print annotate_output
                if annotate_output:
                    obj.annotate_output = True

            for f in module.functions:
                connector = None
                if len(f.params) == 0:
                    connector = ModuleConnector(create_null(), 'value',
                                                f.get_spec('output'))
                elif len(f.params) == 1:
                    p = f.params[0]
                    try:
                        constant = create_constant(p, module)
                        connector = ModuleConnector(constant, 'value',
                                                    f.get_spec('output'))
                    except Exception, e:
                        debug.unexpected_exception(e)
                        err = ModuleError(
                                module,
                                "Uncaught exception creating Constant from "
                                "%r: %s" % (
                                p.strValue,
                                debug.format_exception(e)))
                        errors[i] = err
                        to_delete.append(obj.id)
                else:
                    tupleModule = vistrails.core.interpreter.base.InternalTuple()
                    tupleModule.length = len(f.params)
                    for (j,p) in enumerate(f.params):
                        try:
                            constant = create_constant(p, module)
                            constant.update()
                            connector = ModuleConnector(constant, 'value',
                                                        f.get_spec('output'))
                            tupleModule.set_input_port(j, connector)
                        except Exception, e:
                            debug.unexpected_exception(e)
                            err = ModuleError(
                                    module,
                                    "Uncaught exception creating Constant "
                                    "from %r: %s" % (
                                    p.strValue,
                                    debug.format_exception(e)))
                            errors[i] = err
                            to_delete.append(obj.id)
                    connector = ModuleConnector(tupleModule, 'value',
                                                f.get_spec('output'))
                if connector:
                    obj.set_input_port(f.name, connector, is_method=True)

        # Create the new connections
        for i in conn_added_set:
            persistent_id = conn_map[i]
            conn = self._persistent_pipeline.connections[persistent_id]
            src = self._objects[conn.sourceId]
            dst = self._objects[conn.destinationId]
            self.make_connection(conn, src, dst)

        if self.done_summon_hook:
            self.done_summon_hook(self._persistent_pipeline, self._objects)
        for callable_ in done_summon_hooks:
            callable_(self._persistent_pipeline, self._objects)

        tmp_id_to_module_map = {}
        for i, j in tmp_to_persistent_module_map.iteritems():
            tmp_id_to_module_map[i] = self._objects[j]
        return (tmp_id_to_module_map, tmp_to_persistent_module_map.inverse,
                module_added_set, conn_added_set, to_delete, errors)

    def execute_pipeline(self, pipeline, tmp_id_to_module_map, 
                         persistent_to_tmp_id_map, **kwargs):
        def fetch(name, default):
            return kwargs.pop(name, default)
        controller = fetch('controller', None)
        locator = fetch('locator', None)
        current_version = fetch('current_version', None)
        view = fetch('view', DummyView())
        vistrail_variables = fetch('vistrail_variables', None)
        aliases = fetch('aliases', None)
        params = fetch('params', None)
        extra_info = fetch('extra_info', None)
        logger = fetch('logger', DummyLogController)
        sinks = fetch('sinks', None)
        reason = fetch('reason', None)
        actions = fetch('actions', None)
        module_executed_hook = fetch('module_executed_hook', [])
        done_summon_hooks = fetch('done_summon_hooks', [])
        clean_pipeline = fetch('clean_pipeline', False)
        stop_on_error = fetch('stop_on_error', True)
        parent_exec = fetch('parent_exec', None)

        if len(kwargs) > 0:
            raise VistrailsInternalError('Wrong parameters passed '
                                         'to execute_pipeline: %s' % kwargs)

        # LOGGING SETUP
        def get_remapped_id(id):
            return persistent_to_tmp_id_map[id]

        logging_obj = ViewUpdatingLogController(
                logger=logger,
                view=view,
                remap_id=get_remapped_id,
                ids=pipeline.modules.keys(),
                module_executed_hook=module_executed_hook)

        # PARAMETER CHANGES SETUP
        parameter_changes = []
        def change_parameter(obj, name, value):
            parameter_changes.append((get_remapped_id(obj.id),
                                      name, value))
        def make_change_parameter(obj):
            return lambda *args: change_parameter(obj, *args)

        # Update **all** modules in the current pipeline
        for i, obj in tmp_id_to_module_map.iteritems():
            obj.in_pipeline = True # set flag to indicate in pipeline
            obj.logging = logging_obj
            obj.change_parameter = make_change_parameter(obj)
            
            # Update object pipeline information
            obj.moduleInfo['locator'] = locator
            obj.moduleInfo['version'] = current_version
            obj.moduleInfo['moduleId'] = i
            obj.moduleInfo['pipeline'] = pipeline
            obj.moduleInfo['controller'] = controller
            if extra_info is not None:
                obj.moduleInfo['extra_info'] = extra_info
            if reason is not None:
                obj.moduleInfo['reason'] = reason
            if actions is not None:
                obj.moduleInfo['actions'] = actions

        ## Checking 'sinks' from kwargs to resolve only requested sinks
        # Note that we accept any module in 'sinks', even if it's not actually
        # a sink in the graph
        if sinks is not None:
            persistent_sinks = [tmp_id_to_module_map[sink]
                                for sink in sinks
                                if sink in tmp_id_to_module_map]
        else:
            persistent_sinks = [tmp_id_to_module_map[sink]
                                for sink in pipeline.graph.sinks()]

        self._streams.append(Generator.generators)
        Generator.generators = []

        # Update new sinks
        for obj in persistent_sinks:
            abort = False
            try:
                obj.update()
                continue
            except ModuleWasSuspended:
                continue
            except ModuleHadError:
                pass
            except AbortExecution:
                break
            except ModuleSuspended, ms:
                ms.module.logging.end_update(ms.module, ms,
                                             was_suspended=True)
                continue
            except ModuleErrors, mes:
                for me in mes.module_errors:
                    me.module.logging.end_update(me.module, me)
                    logging_obj.signalError(me.module, me)
                    abort = abort or me.abort
            except ModuleError, me:
                me.module.logging.end_update(me.module, me, me.errorTrace)
                logging_obj.signalError(me.module, me)
                abort = me.abort
            except ModuleBreakpoint, mb:
                mb.module.logging.end_update(mb.module)
                logging_obj.signalError(mb.module, mb)
                abort = True
            if stop_on_error or abort:
                break

        # execute all generators until inputs are exhausted
        # this makes sure branching and multiple sinks are executed correctly
        if not logging_obj.errors and not logging_obj.suspended and \
                                                          Generator.generators:
            result = True
            abort = False
            while result is not None:
                try:
                    for m in Generator.generators:
                        result = m.generator.next()
                    continue
                except AbortExecution:
                    break
                except ModuleErrors, mes:
                    for me in mes.module_errors:
                        me.module.logging.end_update(me.module, me)
                        logging_obj.signalError(me.module, me)
                        abort = abort or me.abort
                except ModuleError, me:
                    me.module.logging.end_update(me.module, me, me.errorTrace)
                    logging_obj.signalError(me.module, me)
                    abort = me.abort
                except ModuleBreakpoint, mb:
                    mb.module.logging.end_update(mb.module)
                    logging_obj.signalError(mb.module, mb)
                    abort = True
                except Exception, e:
                    import traceback
                    traceback.print_exc()
                    abort = True
                if stop_on_error or abort:
                    break

        Generator.generators = self._streams.pop()

        if self.done_update_hook:
            self.done_update_hook(self._persistent_pipeline, self._objects)
                
        # objs, errs, and execs are mappings that use the local ids as keys,
        # as opposed to the persistent ids.
        # They are thus ideal to external consumption.
        objs = {}
        # dict([(i, self._objects[tmp_to_persistent_module_map[i]])
        #              for i in tmp_to_persistent_module_map.keys()])
        errs = {}
        execs = {}
        suspends = {}
        caches = {}

        to_delete = []
        for (tmp_id, obj) in tmp_id_to_module_map.iteritems():
            if clean_pipeline:
                to_delete.append(obj.id)
            objs[tmp_id] = obj
            if obj.id in logging_obj.errors:
                errs[tmp_id] = logging_obj.errors[obj.id]
                if not clean_pipeline:
                    to_delete.append(obj.id)
            executed = False
            if obj.id in logging_obj.executed:
                execs[tmp_id] = logging_obj.executed[obj.id]
                executed = True
            if obj.id in logging_obj.suspended:
                suspends[tmp_id] = logging_obj.suspended[obj.id]
                if not clean_pipeline:
                    to_delete.append(obj.id)
                executed = True
            if obj.id in logging_obj.cached:
                caches[tmp_id] = logging_obj.cached[obj.id]
                executed = True
            if not executed:
                # these modules didn't execute
                execs[tmp_id] = False

        return (to_delete, objs, errs, execs, suspends, caches, parameter_changes)

    def finalize_pipeline(self, pipeline, to_delete, objs, errs, execs,
                          suspended, cached, **kwargs):
        def fetch(name, default):
            return kwargs.pop(name, default)
        reset_computed = fetch('reset_computed', True)
        view = fetch('view', None)

        self.clean_modules(to_delete)

        def dict2set(s):
            return set(k for k, v in s.iteritems() if v)
        if view is not None:
            persistent = set(objs) - (dict2set(errs) | dict2set(execs) |
                                      dict2set(suspended) | dict2set(cached))
            for i in persistent:
                view.set_module_persistent(i)

        if reset_computed:
            for module in self._objects.itervalues():
                module.computed = False

    def execute(self, pipeline, **kwargs):
        """execute(pipeline, **kwargs):

        kwargs:
          controller = fetch('controller', None)
          locator = fetch('locator', None)
          current_version = fetch('current_version', None)
          view = fetch('view', DummyView())
          aliases = fetch('aliases', None)
          params = fetch('params', None)
          extra_info = fetch('extra_info', None)
          logger = fetch('logger', DummyLogController)
          reason = fetch('reason', None)
          actions = fetch('actions', None)
          done_summon_hooks = fetch('done_summon_hooks', [])
          module_executed_hook = fetch('module_executed_hook', [])

        Executes a pipeline using caching. Caching works by reusing
        pipelines directly.  This means that there exists one global
        pipeline whose parts get executed over and over again.

        This function returns a triple of dictionaries (objs, errs, execs).

        objs is a mapping from local ids (the ids in the pipeline) to
        objects **in the persistent pipeline**. Notice, these are not
        the objects inside the passed pipeline, but the objects they
        were mapped to in the persistent pipeline.

        errs is a dictionary from local ids to error messages of modules
        that might have returns errors.

        execs is a dictionary from local ids to boolean values indicating
        whether they were executed or not.

        If modules have no error associated with but were not executed, it
        means they were cached."""

        # Setup named arguments. We don't use named parameters so
        # that positional parameter calls fail earlier
        new_kwargs = {}
        def fetch(name, default):
            new_kwargs[name] = r = kwargs.pop(name, default)
            return r
        controller = fetch('controller', None)
        locator = fetch('locator', None)
        current_version = fetch('current_version', None)
        view = fetch('view', DummyView())
        vistrail_variables = fetch('vistrail_variables', None)
        aliases = fetch('aliases', None)
        params = fetch('params', None)
        extra_info = fetch('extra_info', None)
        logger = fetch('logger', DummyLogController)
        sinks = fetch('sinks', None)
        reason = fetch('reason', None)
        actions = fetch('actions', None)
        done_summon_hooks = fetch('done_summon_hooks', [])
        module_executed_hook = fetch('module_executed_hook', [])
        stop_on_error = fetch('stop_on_error', True)
        parent_exec = fetch('parent_exec', None)

        if len(kwargs) > 0:
            raise VistrailsInternalError('Wrong parameters passed '
                                         'to execute: %s' % kwargs)
        self.clean_non_cacheable_modules()


#         if controller is not None:
#             vistrail = controller.vistrail
#             (pipeline, module_remap) = \
#                 core.db.io.expand_workflow(vistrail, pipeline)
#             new_kwargs['module_remap'] = module_remap
#         else:
#             vistrail = None

        if controller is not None:
            vistrail = controller.vistrail
        else:
            vistrail = None

        logger = logger.start_workflow_execution(
                parent_exec,
                vistrail, pipeline, current_version)
        new_kwargs['logger'] = logger
        self.annotate_workflow_execution(logger, reason, aliases, params)

        res = self.setup_pipeline(pipeline, **new_kwargs)
        modules_added = res[2]
        conns_added = res[3]
        to_delete = res[4]
        errors = res[5]
        if len(errors) == 0:
            res = self.execute_pipeline(pipeline, *(res[:2]), **new_kwargs)
        else:
            res = (to_delete, res[0], errors, {}, {}, {}, [])
            for (i, error) in errors.iteritems():
                view.set_module_error(i, error)
        self.finalize_pipeline(pipeline, *(res[:-1]), **new_kwargs)

        result = InstanceObject(objects=res[1],
                              errors=res[2],
                              executed=res[3],
                              suspended=res[4],
                              parameter_changes=res[6],
                              modules_added=modules_added,
                              conns_added=conns_added)

        logger.finish_workflow_execution(result.errors, suspended=result.suspended)

        return result

    def annotate_workflow_execution(self, logger, reason, aliases, params):
        """annotate_workflow_Execution(logger: LogController, reason:str,
                                        aliases:dict, params:list)-> None
        It will annotate the workflow execution in logger with the reason,
        aliases and params.
        
        """
        d = {}
        d["__reason__"] = reason
        if aliases is not None and isinstance(aliases, dict):
            d["__aliases__"] = pickle.dumps(aliases)
        if params is not None and isinstance(params, list):
            d["__params__"] = pickle.dumps(params)
        logger.insert_workflow_exec_annotations(d)
        
    def add_to_persistent_pipeline(self, pipeline):
        """add_to_persistent_pipeline(pipeline):
        (module_id_map, connection_id_map, modules_added)
        Adds a pipeline to the persistent pipeline of the cached interpreter
        and adds current logging object to each existing module.

        Returns four things: two dictionaries describing the mapping
        of ids from the passed pipeline to the persistent one (the
        first one has the module id mapping, the second one has the
        connection id mapping), a set of all module ids added to the
        persistent pipeline, and a set of all connection ids added to
        the persistent pipeline."""
        module_id_map = Bidict()
        connection_id_map = Bidict()
        modules_added = set()
        connections_added = set()
        pipeline.refresh_signatures()
        # we must traverse vertices in topological sort order
        verts = pipeline.graph.vertices_topological_sort()
        for new_module_id in verts:
            new_sig = pipeline.subpipeline_signature(new_module_id)
            if not self._persistent_pipeline.has_subpipeline_signature(new_sig):
                # Must add module to persistent pipeline
                persistent_module = copy.copy(pipeline.modules[new_module_id])
                persistent_id = self._persistent_pipeline.fresh_module_id()
                persistent_module.id = persistent_id
                self._persistent_pipeline.add_module(persistent_module)
                self._persistent_pipeline.modules[persistent_id]._signature = \
                    base64.b16encode(new_sig).lower()
                module_id_map[new_module_id] = persistent_id
                modules_added.add(new_module_id)
            else:
                i = self._persistent_pipeline \
                        .subpipeline_id_from_signature(new_sig)
                module_id_map[new_module_id] = i
        for connection in pipeline.connections.itervalues():
            new_sig = pipeline.connection_signature(connection.id)
            if not self._persistent_pipeline.has_connection_signature(new_sig):
                # Must add connection to persistent pipeline
                persistent_connection = copy.copy(connection)
                persistent_id = self._persistent_pipeline.fresh_connection_id()
                persistent_connection.id = persistent_id
                persistent_connection.sourceId = module_id_map[
                    connection.sourceId]
                persistent_connection.destinationId = module_id_map[
                    connection.destinationId]
                self._persistent_pipeline.add_connection(persistent_connection)
                connection_id_map[connection.id] = persistent_id
                connections_added.add(connection.id)
            else:
                i = self._persistent_pipeline \
                        .connection_id_from_signature(new_sig)
                connection_id_map[connection.id] = i
        # update persistent signatures
        self._persistent_pipeline.compute_signatures()
        return (module_id_map, connection_id_map,
                modules_added, connections_added)
        
    def find_persistent_entities(self, pipeline):
        """returns a map from a pipeline to the persistent pipeline, 
        assuming those pieces exist"""
        persistent_p = self._persistent_pipeline
        object_map = {}
        module_id_map = {}
        connection_id_map = {}
        pipeline.refresh_signatures()
        # we must traverse vertices in topological sort order
        verts = pipeline.graph.vertices_topological_sort()
        for module_id in verts:
            sig = pipeline.subpipeline_signature(module_id)
            if persistent_p.has_subpipeline_signature(sig):
                i = persistent_p.subpipeline_id_from_signature(sig)
                module_id_map[module_id] = i
                object_map[module_id] = self._objects[i]
            else:
                module_id_map[module_id] = None
                object_map[module_id] = None
        for connection in pipeline.connections.itervalues():
            sig = pipeline.connection_signature(connection.id)
            if persistent_p.has_connection_signature(sig):
                connection_id_map[connection.id] = \
                    persistent_p.connection_id_from_signature(sig)
            else:
                connection_id_map[connection.id] = None
        return (object_map, module_id_map, connection_id_map)

    __instance = None
    @staticmethod
    def get():
        if not CachedInterpreter.__instance:
            CachedInterpreter.__instance = CachedInterpreter()
        return CachedInterpreter.__instance

    @staticmethod
    def cleanup():
        if CachedInterpreter.__instance:
            CachedInterpreter.__instance.clear()
        objs = gc.collect()

    @staticmethod
    def flush():
        if CachedInterpreter.__instance:
            CachedInterpreter.__instance.clear()
            CachedInterpreter.__instance.create()
        objs = gc.collect()

    @staticmethod
    def clear_package(identifier):
        if CachedInterpreter.__instance:
            CachedInterpreter.__instance._clear_package(identifier)

###############################################################################
# Testing

import unittest


class TestCachedInterpreter(unittest.TestCase):

    def test_cache(self):
        from vistrails.core.modules.basic_modules import StandardOutput
        old_compute = StandardOutput.compute
        StandardOutput.compute = lambda s: None

        try:
            from vistrails.core.db.locator import XMLFileLocator
            from vistrails.core.vistrail.controller import VistrailController
            from vistrails.core.db.io import load_vistrail

            """Test if basic caching is working."""
            locator = XMLFileLocator(vistrails.core.system.vistrails_root_directory() +
                                '/tests/resources/dummy.xml')
            (v, abstractions, thumbnails, mashups) = load_vistrail(locator)

            # the controller will take care of upgrades
            controller = VistrailController(v, locator, abstractions,
                                            thumbnails,  mashups)
            p1 = v.getPipeline('int chain')
            n = v.get_version_number('int chain')
            controller.change_selected_version(n)
            controller.flush_delayed_actions()
            p1 = controller.current_pipeline

            view = DummyView()
            interpreter = CachedInterpreter.get()
            result = interpreter.execute(p1,
                                         locator=v,
                                         current_version=n,
                                         view=view,
                                         )
            # to force fresh params
            p2 = v.getPipeline('int chain')
            controller.change_selected_version(n)
            controller.flush_delayed_actions()
            p2 = controller.current_pipeline
            result = interpreter.execute(p2,
                                         locator=v,
                                         current_version=n,
                                         view=view,
                                         )
            self.assertEqual(len(result.modules_added), 1)
        finally:
            StandardOutput.compute = old_compute


if __name__ == '__main__':
    unittest.main()
